

from ..vaex.ml.incubator.pygbm import *
import traitlets

def py_gbm_model(self, features, param, learning_rate=0.1, max_bins=255, max_iter=10, max_leaf_nodes=31, num_round=0, prediction_name='pygbm_prediction', random_state=0, verbose=1):
    obj = vaex.ml.incubator.pygbm.PyGBMModel([])
    obj.fit(self)
    return obj

import vaex.dataset
vaex.dataset.Dataset.ml._add(py_gbm_model=py_gbm_model)

def __init__(self, features=traitlets.Undefined, learning_rate=0.1, max_bins=255, max_iter=10, max_leaf_nodes=31, num_round=0, param=traitlets.Undefined, prediction_name='pygbm_prediction', random_state=0, verbose=1):
    """
    

    """
    given_kwargs = {key:value for key, value in dict(features=features, learning_rate=learning_rate, max_bins=max_bins, max_iter=max_iter, max_leaf_nodes=max_leaf_nodes, num_round=num_round, param=param, prediction_name=prediction_name, random_state=random_state, verbose=verbose).items() if value is not traitlets.Undefined}

    super(vaex.ml.incubator.pygbm.PyGBMModel, self).__init__(**given_kwargs)

PyGBMModel.__init__ = __init__
PyGBMModel.__signature__ = __init__
del __init__
    




from ..vaex.ml.linear_model import *
import traitlets

def logistic_regression(self, features, _sk_params=None, binned=True, fit_intercept=True, prediction_name='logit_prediction', shape=64):
    obj = vaex.ml.linear_model.LogisticRegression([])
    obj.fit(self)
    return obj

import vaex.dataset
vaex.dataset.Dataset.ml._add(logistic_regression=logistic_regression)

def __init__(self, _sk_params=None, binned=True, features=traitlets.Undefined, fit_intercept=True, prediction_name='logit_prediction', shape=64):
    """
    

    """
    given_kwargs = {key:value for key, value in dict(_sk_params=_sk_params, binned=binned, features=features, fit_intercept=fit_intercept, prediction_name=prediction_name, shape=shape).items() if value is not traitlets.Undefined}

    super(vaex.ml.linear_model.LogisticRegression, self).__init__(**given_kwargs)

LogisticRegression.__init__ = __init__
LogisticRegression.__signature__ = __init__
del __init__
    


from ..vaex.ml.linear_model import *
import traitlets

def linear_regression(self, features, _sk_params=None, binned=True, fit_intercept=True, prediction_name='linear_prediction', shape=64):
    obj = vaex.ml.linear_model.LinearRegression([])
    obj.fit(self)
    return obj

import vaex.dataset
vaex.dataset.Dataset.ml._add(linear_regression=linear_regression)

def __init__(self, _sk_params=None, binned=True, features=traitlets.Undefined, fit_intercept=True, prediction_name='linear_prediction', shape=64):
    """
    

    """
    given_kwargs = {key:value for key, value in dict(_sk_params=_sk_params, binned=binned, features=features, fit_intercept=fit_intercept, prediction_name=prediction_name, shape=shape).items() if value is not traitlets.Undefined}

    super(vaex.ml.linear_model.LinearRegression, self).__init__(**given_kwargs)

LinearRegression.__init__ = __init__
LinearRegression.__signature__ = __init__
del __init__
    


import vaex.dataset
from vaex.serialize import register
import numpy as np
from . import generate
from .state import HasState
from traitlets import Dict, Unicode, List, Int, CFloat, CBool, Any, Tuple

help_features = 'List of features to transform.'
help_prefix = 'Prefix for the names of the transformed features.'


def dot_product(a, b):
    products = ['%s * %s' % (ai, bi) for ai, bi in zip(a, b)]
    return ' + '.join(products)


@register
class StateTransfer(HasState):
    state = Dict()

    def transform(self, dataset):
        copy = dataset.copy()
        self.state = dict(self.state, active_range=[copy._index_start, copy._index_end])
        copy.state_set(self.state)
        return copy


class Transformer(HasState):
    ''' Parent class for all of the transformers.
    Currently it implements only the fit_predict() method.
    '''

    def fit_transform(self, dataset):
        '''Fit and apply the transformer to the supplied dataset.

        Parameters:
        -----------
        :param dataset: A vaex dataset.

        Returns:
        --------
        :param copy: A shallow copy of the dataset that includes the transformations.
        '''

        self.fit(dataset=dataset)
        return self.transform(dataset=dataset)


@register
@generate.register
class PCA(Transformer):
    '''Transform a set of features using a Principal Component Analysis.

    Example
    -------

    >>> import vaex.ml
    >>> df = vaex.ml.datasets.load_iris()
    >>> features = ['sepal_width', 'petal_length', 'sepal_length', 'petal_width']
    >>> df_train, df_test = vaex.ml.train_test_split(df)
    >>> pca = vaex.ml.PCA(features=features, n_components=3)
    >>> df_train = pca.fit_transform(df_train)
    >>> df_test = pca.transform(df_test)
    '''

    # title = Unicode(default_value='PCA', read_only=True).tag(ui='HTML')
    features = List(Unicode(), help=help_features).tag(ui='SelectMultiple')
    n_components = Int(default_value=2, help='Number of components to retain.').tag(ui='IntText')
    prefix = Unicode(default_value="PCA_", help=help_prefix)
    eigen_vectors_ = List(List(CFloat()), help='The eigen vectors corresponding to each feature').tag(output=True)
    eigen_values_ = List(CFloat(), help='The eigen values that correspond to each feature.').tag(output=True)
    means_ = List(CFloat(), help='The mean of each feature').tag(output=True)

    def fit(self, dataset, column_names=None, progress=False):
        '''
        Fit the PCA model to the dataset.

        Parameters:
        -----------
        :param dataset: A vaex dataset.
        :param columns: Deprecated and should be removed.
        :param progress: bool, if True, display a progress bar of the fitting process.

        Returns:
        --------
        self
        '''

        assert self.n_components <= len(self.features), 'cannot have more components than features'
        C = dataset.cov(self.features, progress=progress)
        eigen_values, eigen_vectors = np.linalg.eigh(C)
        indices = np.argsort(eigen_values)[::-1]
        self.means_ = dataset.mean(self.features, progress=progress).tolist()
        self.eigen_vectors_ = eigen_vectors[:, indices].tolist()
        self.eigen_values_ = eigen_values[indices].tolist()

    def transform(self, dataset, n_components=None):
        '''
        Apply the PCA transformation to the dataset.

        Parameters:
        -----------
        :param dataset: A vaex dataset.
        :param n_components: The number of components to retain.

        Returns:
        --------
        :return copy: A shallow copy of the dataset that includes the PCA components.
        '''

        n_components = n_components or self.n_components
        copy = dataset.copy()
        name_prefix_offset = 0
        eigen_vectors = np.array(self.eigen_vectors_)
        while self.prefix + str(name_prefix_offset) in copy.get_column_names(virtual=True, strings=True):
            name_prefix_offset += 1

        expressions = [copy[feature]-mean for feature, mean in zip(self.features, self.means_)]
        for i in range(n_components):
            v = eigen_vectors[:, i]
            expr = dot_product(expressions, v)
            name = self.prefix + str(i + name_prefix_offset)
            copy[name] = expr
        return copy


@register
@generate.register
class LabelEncoder(Transformer):
    '''Encode categorical columns with integer values between 0 and num_classes-1.

    Example
    -------

    >>> import vaex.ml
    >>> df = vaex.ml.datasets.load_titanic()
    >>> df_train, df_test = vaex.ml.train_test_split(df)
    >>> encoder = vaex.ml.LabelEncoder(features=['sex', 'embarked'])
    >>> df_train = encoder.fit_transform(df_train)
    >>> df_test = encoder.transform(df_test)
    '''

    # title = Unicode(default_value='Label Encoder', read_only=True).tag(ui='HTML')
    features = List(Unicode(), help=help_features).tag(ui='SelectMultiple')
    prefix = Unicode(default_value=help_prefix, help="").tag(ui='Text')
    labels_ = List(List(), help='The encoded labels of each feature.').tag(output=True)

    def fit(self, dataset):
        '''
        Fit LabelEncoder to the dataset.

        Parameters:
        -----------
        :param dataset: A vaex dataset.

        Returns:
        --------
        self
        '''

        labels = []
        for i in self.features:
            labels.append(np.unique(dataset.evaluate(i)).tolist())
        self.labels_ = labels

    def transform(self, dataset):
        '''
        Transform a dataset with a fitted LabelEncoder.

        Parameters:
        -----------
        :param dataset: A vaex dataset.

        Returns:
        --------
        :return copy: A shallow copy of the dataset that includes the encodings.
        '''

        copy = dataset.copy()
        for i, v in enumerate(self.features):
            name = self.prefix + v
            labels = np.unique(dataset.evaluate(v))
            if len(np.intersect1d(labels, self.labels_[i])) < len(labels):
                diff = np.setdiff1d(labels, self.labels_[i])
                raise ValueError("%s contains previously unseen labels: %s" % (v, str(diff)))
            # copy[name] = np.searchsorted(self.labels[i], v)
            copy.add_virtual_column(name, 'searchsorted({x}, {v})'.format(x=self.labels_[i], v=v))
        return copy


@register
@generate.register
class OneHotEncoder(Transformer):
    '''Encode categorical columns according ot the One-Hot scheme.

    Example
    -------

    >>> import vaex.ml
    >>> df = vaex.from_arrays(color=['red', 'green', 'green', 'blue', 'red'])
    >>> df
     #  color
     0  red
     1  green
     2  green
     3  blue
     4  red
    >>> encoder = vaex.ml.OneHotEncoder(features=['color'])
    >>> encoder.fit_transform(df)
     #  color      color_blue    color_green    color_red
     0  red                 0              0            1
     1  green               0              1            0
     2  green               0              1            0
     3  blue                1              0            0
     4  red                 0              0            1

    '''

    # title = Unicode(default_value='One-Hot Encoder', read_only=True).tag(ui='HTML')
    features = List(Unicode(), help=help_features).tag(ui='SelectMultiple')
    prefix = Unicode(default_value='', help=help_prefix).tag(ui='Text')
    one = Any(1, help='Value to encode when a category is present.')
    zero = Any(0, help='Value to encode when category is absent.')
    uniques_ = List(List(), help='The unique elements found in each feature.').tag(output=True)

    def fit(self, dataset):
        '''Fit OneHotEncoder to the dataset.

        :param dataset: A vaex dataset.
        '''

        uniques = []
        for i in self.features:
            expression = vaex.dataset._ensure_strings_from_expressions(i)
            unique = dataset.unique(expression)
            unique = np.sort(unique)  # this can/should be optimized with @delay
            uniques.append(unique.tolist())
        self.uniques_ = uniques

    def transform(self, dataset):
        '''Transform a dataset with a fitted OneHotEncoder.

        :param dataset: A vaex dataset.
        :return: A shallow copy of the dataset that includes the encodings.
        '''

        copy = dataset.copy()
        # for each feature, add a virtual column for each unique entry
        for i, feature in enumerate(self.features):
            for j, value in enumerate(self.uniques_[i]):
                column_name = self.prefix + feature + '_' + str(value)
                copy.add_virtual_column(column_name, 'where({feature} == {value}, {one}, {zero})'.format(
                                        feature=feature, value=repr(value), one=self.one, zero=self.zero))
        return copy


@register
@generate.register
class StandardScaler(Transformer):
    '''Standardize features by removing thir mean and scaling them to unit variance.

    Example:
    --------

    >>> import vaex.ml
    >>> df = vaex.ml.datasets.load_iris()
    >>> features = ['sepal_width', 'petal_length', 'sepal_length', 'petal_width']
    >>> df_train, df_test = vaex.ml.train_test_split(df)
    >>> scaler = vaex.ml.StandardScaler(features=features, with_mean=True, with_std=True)
    >>> df_train = scaler.fit_transform(df_train)
    >>> df_test = scaler.transform(df_test)
    '''

    # title = Unicode(default_value='Standard Scaler', read_only=True).tag(ui='HTML')
    features = List(Unicode(), help=help_features).tag(ui='SelectMultiple')
    prefix = Unicode(default_value="standard_scaled_", help=help_prefix).tag(ui='Text')
    with_mean = CBool(default_value=True, help='If True, remove the mean from each feature.').tag(ui='Checkbox')
    with_std = CBool(default_value=True, help='If True, scale each feature to unit variance.').tag(ui='Checkbox')
    mean_ = List(CFloat(), help='The mean of each feature').tag(output=True)
    std_ = List(CFloat(), help='The standard deviation of each feature.').tag(output=True)

    def fit(self, dataset):
        '''
        Fit StandardScaler to the dataset.

        Parameters:
        -----------
        :param dataset: A vaex dataset.

        Returns:
        --------
        self
        '''

        mean = dataset.mean(self.features, delay=True)
        std = dataset.std(self.features, delay=True)

        @vaex.delayed
        def assign(mean, std):
            self.mean_ = mean.tolist()
            self.std_ = std.tolist()

        assign(mean, std)
        dataset.executor.execute()

    def transform(self, dataset):
        '''
        Transform a dataset with a fitted StandardScaler.

        Parameters:
        -----------
        :param dataset: A vaex dataset.

        Returns:
        --------
        :return copy: a shallow copy of the dataset that includes the scaled features.
        '''

        copy = dataset.copy()
        for i in range(len(self.features)):
            name = self.prefix+self.features[i]
            expression = copy[self.features[i]]
            if self.with_mean:
                expression = expression - self.mean_[i]
            if self.with_std:
                expression = expression / self.std_[i]
            copy[name] = expression
        return copy


@register
@generate.register
class MinMaxScaler(Transformer):
    '''Will scale a set of features to a given range.

    Example:
    --------

    >>> import vaex.ml
    >>> df = vaex.ml.datasets.load_iris()
    >>> features = ['sepal_width', 'petal_length', 'sepal_length', 'petal_width']
    >>> df_train, df_test = vaex.ml.train_test_split(df)
    >>> scaler = vaex.ml.MinMaxScaler(features=features, feature_range=(0, 1))
    >>> df_train = scaler.fit_transform(df_train)
    >>> df_test = scaler.transform(df_test)
    '''

    # title = Unicode(default_value='MinMax Scaler', read_only=True).tag(ui='HTML')
    features = List(Unicode(), help=help_features).tag(ui='SelectMultiple')
    feature_range = Tuple(default_value=(0, 1), help='The range the features are scaled to.').tag().tag(ui='FloatRangeSlider')
    prefix = Unicode(default_value="minmax_scaled_", help=help_prefix).tag(ui='Text')
    fmax_ = List(CFloat(), help='The minimum value of a feature.').tag(output=True)
    fmin_ = List(CFloat(), help='The maximum value of a feature.').tag(output=True)

    def fit(self, dataset):
        '''
        Fit MinMaxScaler to the dataset.

        Parameters:
        -----------
        :param dataset: A vaex dataset.

        Returns:
        --------
        self
        '''

        assert len(self.feature_range) == 2, 'feature_range must have 2 elements only'
        minmax = dataset.minmax(self.features)
        self.fmin_ = minmax[:, 0].tolist()
        self.fmax_ = minmax[:, 1].tolist()

    def transform(self, dataset):
        '''
        Transform a dataset with a fitted MinMaxScaler.

        Parameters:
        -----------
        :param dataset: A vaex dataset.

        Returns:
        --------
        :return copy: a shallow copy of the dataset that includes the scaled features.
        '''

        copy = dataset.copy()

        for i in range(len(self.features)):
            name = self.prefix + self.features[i]
            a = self.feature_range[0]
            b = self.feature_range[1]
            expr = copy[self.features[i]]
            expr = (b-a)*(expr-self.fmin_[i])/(self.fmax_[i]-self.fmin_[i]) + a
            copy[name] = expr
        return copy


@register
@generate.register
class MaxAbsScaler(Transformer):
    ''' Scale features by their maximum absolute value.

    Example:
    --------

    >>> import vaex.ml
    >>> df = vaex.ml.datasets.load_iris()
    >>> features = ['sepal_width', 'petal_length', 'sepal_length', 'petal_width']
    >>> df_train, df_test = vaex.ml.train_test_split(df)
    >>> scaler = vaex.ml.MaxAbsScaler(features=features)
    >>> df_train = scaler.fit_transform(df_train)
    >>> df_test = scaler.transform(df_test)
    '''

    features = List(Unicode(), help=help_features).tag(ui='SelectMultiple')
    prefix = Unicode(default_value="absmax_scaled_", help=help_prefix).tag(ui='Text')
    absmax_ = List(CFloat(), help='Tha maximum absolute value of a feature.').tag(output=True)

    def fit(self, dataset):
        '''
        Fit MinMaxScaler to the dataset.

        Parameters:
        -----------
        :param dataset: A vaex dataset.

        Returns:
        --------
        self
        '''

        absmax = dataset.max(['abs(%s)' % k for k in self.features]).tolist()
        # Check if the absmax_ value is 0, in which case replace with 1
        self.absmax_ = [value if value != 0 else 1 for value in absmax]

    def transform(self, dataset):
        '''
        Transform a dataset with a fitted MaxAbsScaler.

        Parameters:
        -----------
        :param dataset: A vaex dataset.

        Returns:
        --------
        :return copy: a shallow copy of the dataset that includes the scaled features.
        '''

        copy = dataset.copy()
        for i in range(len(self.features)):
            name = self.prefix + self.features[i]
            expr = copy[self.features[i]]
            expr = expr / self.absmax_[i]
            copy[name] = expr
        return copy


@register
@generate.register
class RobustScaler(Transformer):
    ''' The RobustScaler removes the median and scales the data according to a
    given percentile range. By default, the scaling is done between the 25th and
    the 75th percentile. Centering and scaling happens independently for each
    feature (column).

    Example:
    --------

    >>> import vaex.ml
    >>> df = vaex.ml.datasets.load_iris()
    >>> features = ['sepal_width', 'petal_length', 'sepal_length', 'petal_width']
    >>> df_train, df_test = vaex.ml.train_test_split(df)
    >>> scaler = vaex.ml.RobustScaler(features=features, percentile_range=(25, 75))
    >>> df_train = scaler.fit_transform(df_train)
    >>> df_test = scaler.transform(df_test)
    '''

    features = List(Unicode(), help=help_features).tag(ui='SelectMultiple')
    with_centering = CBool(default_value=True, help='If True, remove the median.').tag(ui='Checkbox')
    with_scaling = CBool(default_value=True, help='If True, scale each feature between the specified percentile range.').tag(ui='Checkbox')
    percentile_range = Tuple(default_value=(25, 75), help='The percentile range to which to scale each feature to.').tag().tag(ui='FloatRangeSlider')
    prefix = Unicode(default_value="robust_scaled_", help=help_prefix).tag(ui='Text')
    center_ = List(CFloat(), default_value=None, help='The median of each feature.').tag(output=True)
    scale_ = List(CFloat(), default_value=None, help='The percentile range for each feature.').tag(output=True)

    def fit(self, dataset):
        '''
        Fit RobustScaler to the dataset.

        Parameters:
        -----------
        :param dataset: A vaex dataset.

        Returns:
        --------
        self
        '''

        # check the quantile range
        q_min, q_max = self.percentile_range
        if not 0 <= q_min <= q_max <= 100:
            raise ValueError('Invalid percentile range: %s' % (str(self.percentile_range)))

        if self.with_centering:
            self.center_ = dataset.percentile_approx(expression=self.features, percentage=50).tolist()

        if self.with_scaling:
            self.scale_ = (dataset.percentile_approx(expression=self.features, percentage=q_max) - dataset.percentile_approx(expression=self.features, percentage=q_min)).tolist()

    def transform(self, dataset):
        '''
        Transform a dataset with a fitted RobustScaler.

        Parameters:
        -----------
        :param dataset: A vaex dataset.

        Returns:
        --------
        :return copy: a shallow copy of the dataset that includes the scaled features.
        '''

        copy = dataset.copy()
        for i in range(len(self.features)):
            name = self.prefix+self.features[i]
            expr = copy[self.features[i]]
            if self.with_centering:
                expr = expr - self.center_[i]
            if self.with_scaling:
                expr = expr / self.scale_[i]
            copy[name] = expr
        return copy
try:
    from .autogen import transformations
    del transformations
except ImportError:
    pass





import vaex.ml.ui
import traitlets

def widgetize(self, _display_callbacks, _dom_classes, _msg_callbacks, _property_lock, _states_to_send, children, comm, keys, layout, log, _model_module='@jupyter-widgets/controls', _model_module_version='1.4.0', _model_name='VBoxModel', _view_count=None, _view_module='@jupyter-widgets/controls', _view_module_version='1.4.0', _view_name='VBoxView', box_style=''):
    obj = vaex.ml.ui.Widgetize([{'name': '_dom_classes', 'help': 'CSS classes applied to widget DOM element'}, {'name': '_view_count', 'help': 'EXPERIMENTAL: The number of views of the model displayed in the frontend. This attribute is experimental and may change or be removed in the future. None signifies that views will not be tracked. Set this to 0 to start tracking view creation/deletion.'}, {'name': 'box_style', 'help': 'Use a predefined styling for the box.'}, {'name': 'children', 'help': 'List of widget children'}, {'name': 'keys', 'help': 'The traits which are synced.'}])
    obj.fit(self)
    return obj

import vaex.dataset
vaex.dataset.Dataset.ml._add(widgetize=widgetize)

def __init__(self, _display_callbacks=traitlets.Undefined, _dom_classes=traitlets.Undefined, _model_module='@jupyter-widgets/controls', _model_module_version='1.4.0', _model_name='VBoxModel', _msg_callbacks=traitlets.Undefined, _property_lock=traitlets.Undefined, _states_to_send=traitlets.Undefined, _view_count=None, _view_module='@jupyter-widgets/controls', _view_module_version='1.4.0', _view_name='VBoxView', box_style='', children=traitlets.Undefined, comm=traitlets.Undefined, keys=traitlets.Undefined, layout=traitlets.Undefined, log=traitlets.Undefined):
    """
    

:param _dom_classes: CSS classes applied to widget DOM element
:param _view_count: EXPERIMENTAL: The number of views of the model displayed in the frontend. This attribute is experimental and may change or be removed in the future. None signifies that views will not be tracked. Set this to 0 to start tracking view creation/deletion.
:param box_style: Use a predefined styling for the box.
:param children: List of widget children
:param keys: The traits which are synced.
    """
    super(vaex.ml.ui.Widgetize, self).__init__(_display_callbacks=_display_callbacks, _dom_classes=_dom_classes, _model_module=_model_module, _model_module_version=_model_module_version, _model_name=_model_name, _msg_callbacks=_msg_callbacks, _property_lock=_property_lock, _states_to_send=_states_to_send, _view_count=_view_count, _view_module=_view_module, _view_module_version=_view_module_version, _view_name=_view_name, box_style=box_style, children=children, comm=comm, keys=keys, layout=layout, log=log)

vaex.ml.ui.Widgetize.__init__ = __init__
vaex.ml.ui.Widgetize.__signature__ = __init__
del __init__
    


import vaex.ml.transformations
import traitlets

def pca(self, features, n_components=2, prefix='PCA_'):
    obj = vaex.ml.transformations.PCA([])
    obj.fit(self)
    return obj

import vaex.dataset
vaex.dataset.Dataset.ml._add(pca=pca)

def __init__(self, features=traitlets.Undefined, n_components=2, prefix='PCA_'):
    """
    

    """
    super(vaex.ml.transformations.PCA, self).__init__(features=features, n_components=n_components, prefix=prefix)

vaex.ml.transformations.PCA.__init__ = __init__
vaex.ml.transformations.PCA.__signature__ = __init__
del __init__
    


import vaex.ml.transformations
import traitlets

def label_encoder(self, features, prefix='label_encoded_'):
    obj = vaex.ml.transformations.LabelEncoder([])
    obj.fit(self)
    return obj

import vaex.dataset
vaex.dataset.Dataset.ml._add(label_encoder=label_encoder)

def __init__(self, features=traitlets.Undefined, prefix='label_encoded_'):
    """
    

    """
    super(vaex.ml.transformations.LabelEncoder, self).__init__(features=features, prefix=prefix)

vaex.ml.transformations.LabelEncoder.__init__ = __init__
vaex.ml.transformations.LabelEncoder.__signature__ = __init__
del __init__
    


import vaex.ml.transformations
import traitlets

def one_hot_encoder(self, features, one=1, prefix='', zero=0):
    obj = vaex.ml.transformations.OneHotEncoder([{'name': 'features', 'help': 'List of features to one hot encode'}, {'name': 'one', 'help': 'Value to use for existing values'}, {'name': 'prefix', 'help': 'Prefix to add to the columns'}, {'name': 'zero', 'help': 'Value to use for missing values'}])
    obj.fit(self)
    return obj

import vaex.dataset
vaex.dataset.Dataset.ml._add(one_hot_encoder=one_hot_encoder)

def __init__(self, features=traitlets.Undefined, one=1, prefix='', zero=0):
    """
    

:param features: List of features to one hot encode
:param one: Value to use for existing values
:param prefix: Prefix to add to the columns
:param zero: Value to use for missing values
    """
    super(vaex.ml.transformations.OneHotEncoder, self).__init__(features=features, one=one, prefix=prefix, zero=zero)

vaex.ml.transformations.OneHotEncoder.__init__ = __init__
vaex.ml.transformations.OneHotEncoder.__signature__ = __init__
del __init__
    


import vaex.ml.transformations
import traitlets

def standard_scaler(self, features, prefix='standard_scaled_', with_mean=True, with_std=True):
    obj = vaex.ml.transformations.StandardScaler([])
    obj.fit(self)
    return obj

import vaex.dataset
vaex.dataset.Dataset.ml._add(standard_scaler=standard_scaler)

def __init__(self, features=traitlets.Undefined, prefix='standard_scaled_', with_mean=True, with_std=True):
    """
    

    """
    super(vaex.ml.transformations.StandardScaler, self).__init__(features=features, prefix=prefix, with_mean=with_mean, with_std=with_std)

vaex.ml.transformations.StandardScaler.__init__ = __init__
vaex.ml.transformations.StandardScaler.__signature__ = __init__
del __init__
    


import vaex.ml.transformations
import traitlets

def min_max_scaler(self, feature_range, features, prefix='minmax_scaled_'):
    obj = vaex.ml.transformations.MinMaxScaler([])
    obj.fit(self)
    return obj

import vaex.dataset
vaex.dataset.Dataset.ml._add(min_max_scaler=min_max_scaler)

def __init__(self, feature_range=traitlets.Undefined, features=traitlets.Undefined, prefix='minmax_scaled_'):
    """
    

    """
    super(vaex.ml.transformations.MinMaxScaler, self).__init__(feature_range=feature_range, features=features, prefix=prefix)

vaex.ml.transformations.MinMaxScaler.__init__ = __init__
vaex.ml.transformations.MinMaxScaler.__signature__ = __init__
del __init__
    





from ..vaex.ml.incubator.annoy import *
import traitlets

def annoy_model(self, features, metric='euclidean', n_neighbours=10, n_trees=10, predcition_name='annoy_prediction', prediction_name='annoy_prediction', search_k=-1):
    obj = vaex.ml.incubator.annoy.ANNOYModel([{'name': 'features', 'help': 'List of features to use.'}, {'name': 'metric', 'help': 'Metric to use for distance calculations'}, {'name': 'n_neighbours', 'help': 'Now many neighbours'}, {'name': 'n_trees', 'help': 'Number of trees to build.'}, {'name': 'predcition_name', 'help': 'Output column name for the neighbours when transforming a dataset'}, {'name': 'prediction_name', 'help': 'Output column name for the neighbours when transforming a dataset'}, {'name': 'search_k', 'help': 'Jovan?'}])
    obj.fit(self)
    return obj

import vaex.dataset
vaex.dataset.Dataset.ml._add(annoy_model=annoy_model)

def __init__(self, features=traitlets.Undefined, metric='euclidean', n_neighbours=10, n_trees=10, predcition_name='annoy_prediction', prediction_name='annoy_prediction', search_k=-1):
    """
    

:param features: List of features to use.
:param metric: Metric to use for distance calculations
:param n_neighbours: Now many neighbours
:param n_trees: Number of trees to build.
:param predcition_name: Output column name for the neighbours when transforming a dataset
:param prediction_name: Output column name for the neighbours when transforming a dataset
:param search_k: Jovan?
    """
    given_kwargs = {key:value for key, value in dict(features=features, metric=metric, n_neighbours=n_neighbours, n_trees=n_trees, predcition_name=predcition_name, prediction_name=prediction_name, search_k=search_k).items() if value is not traitlets.Undefined}

    super(vaex.ml.incubator.annoy.ANNOYModel, self).__init__(**given_kwargs)

ANNOYModel.__init__ = __init__
ANNOYModel.__signature__ = __init__
del __init__
    




from ..vaex.ml.cluster import *
import traitlets

def k_means(self, cluster_centers, features, inertia=None, init='random', max_iter=300, n_clusters=2, n_init=1, prediction_label='prediction_kmeans', random_state=None, verbose=False):
    obj = vaex.ml.cluster.KMeans([{'name': 'cluster_centers', 'help': 'Coordinates of cluster centers.'}, {'name': 'features', 'help': 'List of features to cluster.'}, {'name': 'inertia', 'help': 'Sum of squared distances of samples to their closest cluster center.'}, {'name': 'init', 'help': 'Method for initializing the centroids.'}, {'name': 'max_iter', 'help': 'Maximum number of iterations of the KMeans algorithm for a single run.'}, {'name': 'n_clusters', 'help': 'Number of clusters to form.'}, {'name': 'n_init', 'help': 'Number of centroid initializations.                                                    The KMeans algorithm will be run for each initialization,                                                    and the final results will be the best output of the n_init                                                    consecutive runs in terms of inertia.'}, {'name': 'prediction_label', 'help': 'The name of the virtual column that houses the cluster labels for each point.'}, {'name': 'random_state', 'help': 'Random number generation for centroid initialization.                                                                              If an int is specified, the randomness becomes deterministic.'}, {'name': 'verbose', 'help': 'If True, enable verbosity mode.'}])
    obj.fit(self)
    return obj

import vaex.dataset
vaex.dataset.Dataset.ml._add(k_means=k_means)

def __init__(self, cluster_centers=traitlets.Undefined, features=traitlets.Undefined, inertia=None, init='random', max_iter=300, n_clusters=2, n_init=1, prediction_label='prediction_kmeans', random_state=None, verbose=False):
    """
    

:param cluster_centers: Coordinates of cluster centers.
:param features: List of features to cluster.
:param inertia: Sum of squared distances of samples to their closest cluster center.
:param init: Method for initializing the centroids.
:param max_iter: Maximum number of iterations of the KMeans algorithm for a single run.
:param n_clusters: Number of clusters to form.
:param n_init: Number of centroid initializations.                                                    The KMeans algorithm will be run for each initialization,                                                    and the final results will be the best output of the n_init                                                    consecutive runs in terms of inertia.
:param prediction_label: The name of the virtual column that houses the cluster labels for each point.
:param random_state: Random number generation for centroid initialization.                                                                              If an int is specified, the randomness becomes deterministic.
:param verbose: If True, enable verbosity mode.
    """
    given_kwargs = {key:value for key, value in dict(cluster_centers=cluster_centers, features=features, inertia=inertia, init=init, max_iter=max_iter, n_clusters=n_clusters, n_init=n_init, prediction_label=prediction_label, random_state=random_state, verbose=verbose).items() if value is not traitlets.Undefined}

    super(vaex.ml.cluster.KMeans, self).__init__(**given_kwargs)

KMeans.__init__ = __init__
KMeans.__signature__ = __init__
del __init__
    




from ..vaex.ml.transformations import *
import traitlets

def pca(self, features, n_components=2, prefix='PCA_'):
    obj = vaex.ml.transformations.PCA([{'name': 'features', 'help': 'List of features to transform.'}, {'name': 'n_components', 'help': 'Number of components to retain.'}, {'name': 'prefix', 'help': 'Prefix for the names of the transformed features.'}])
    obj.fit(self)
    return obj

import vaex.dataset
vaex.dataset.Dataset.ml._add(pca=pca)

def __init__(self, features=traitlets.Undefined, n_components=2, prefix='PCA_'):
    """
    

:param features: List of features to transform.
:param n_components: Number of components to retain.
:param prefix: Prefix for the names of the transformed features.
    """
    given_kwargs = {key:value for key, value in dict(features=features, n_components=n_components, prefix=prefix).items() if value is not traitlets.Undefined}

    super(vaex.ml.transformations.PCA, self).__init__(**given_kwargs)

PCA.__init__ = __init__
PCA.__signature__ = __init__
del __init__
    


from ..vaex.ml.transformations import *
import traitlets

def label_encoder(self, features, prefix='Prefix for the names of the transformed features.'):
    obj = vaex.ml.transformations.LabelEncoder([{'name': 'features', 'help': 'List of features to transform.'}, {'name': 'prefix', 'help': ''}])
    obj.fit(self)
    return obj

import vaex.dataset
vaex.dataset.Dataset.ml._add(label_encoder=label_encoder)

def __init__(self, features=traitlets.Undefined, prefix='Prefix for the names of the transformed features.'):
    """
    

:param features: List of features to transform.
:param prefix: 
    """
    given_kwargs = {key:value for key, value in dict(features=features, prefix=prefix).items() if value is not traitlets.Undefined}

    super(vaex.ml.transformations.LabelEncoder, self).__init__(**given_kwargs)

LabelEncoder.__init__ = __init__
LabelEncoder.__signature__ = __init__
del __init__
    


from ..vaex.ml.transformations import *
import traitlets

def one_hot_encoder(self, features, one=1, prefix='', zero=0):
    obj = vaex.ml.transformations.OneHotEncoder([{'name': 'features', 'help': 'List of features to transform.'}, {'name': 'one', 'help': 'Value to encode when a category is present.'}, {'name': 'prefix', 'help': 'Prefix for the names of the transformed features.'}, {'name': 'zero', 'help': 'Value to encode when category is absent.'}])
    obj.fit(self)
    return obj

import vaex.dataset
vaex.dataset.Dataset.ml._add(one_hot_encoder=one_hot_encoder)

def __init__(self, features=traitlets.Undefined, one=1, prefix='', zero=0):
    """
    

:param features: List of features to transform.
:param one: Value to encode when a category is present.
:param prefix: Prefix for the names of the transformed features.
:param zero: Value to encode when category is absent.
    """
    given_kwargs = {key:value for key, value in dict(features=features, one=one, prefix=prefix, zero=zero).items() if value is not traitlets.Undefined}

    super(vaex.ml.transformations.OneHotEncoder, self).__init__(**given_kwargs)

OneHotEncoder.__init__ = __init__
OneHotEncoder.__signature__ = __init__
del __init__
    


from ..vaex.ml.transformations import *
import traitlets

def standard_scaler(self, features, prefix='standard_scaled_', with_mean=True, with_std=True):
    obj = vaex.ml.transformations.StandardScaler([{'name': 'features', 'help': 'List of features to transform.'}, {'name': 'prefix', 'help': 'Prefix for the names of the transformed features.'}, {'name': 'with_mean', 'help': 'If True, remove the mean from each feature.'}, {'name': 'with_std', 'help': 'If True, scale each feature to unit variance.'}])
    obj.fit(self)
    return obj

import vaex.dataset
vaex.dataset.Dataset.ml._add(standard_scaler=standard_scaler)

def __init__(self, features=traitlets.Undefined, prefix='standard_scaled_', with_mean=True, with_std=True):
    """
    

:param features: List of features to transform.
:param prefix: Prefix for the names of the transformed features.
:param with_mean: If True, remove the mean from each feature.
:param with_std: If True, scale each feature to unit variance.
    """
    given_kwargs = {key:value for key, value in dict(features=features, prefix=prefix, with_mean=with_mean, with_std=with_std).items() if value is not traitlets.Undefined}

    super(vaex.ml.transformations.StandardScaler, self).__init__(**given_kwargs)

StandardScaler.__init__ = __init__
StandardScaler.__signature__ = __init__
del __init__
    


from ..vaex.ml.transformations import *
import traitlets

def min_max_scaler(self, feature_range, features, prefix='minmax_scaled_'):
    obj = vaex.ml.transformations.MinMaxScaler([{'name': 'feature_range', 'help': 'The range the features are scaled to.'}, {'name': 'features', 'help': 'List of features to transform.'}, {'name': 'prefix', 'help': 'Prefix for the names of the transformed features.'}])
    obj.fit(self)
    return obj

import vaex.dataset
vaex.dataset.Dataset.ml._add(min_max_scaler=min_max_scaler)

def __init__(self, feature_range=traitlets.Undefined, features=traitlets.Undefined, prefix='minmax_scaled_'):
    """
    

:param feature_range: The range the features are scaled to.
:param features: List of features to transform.
:param prefix: Prefix for the names of the transformed features.
    """
    given_kwargs = {key:value for key, value in dict(feature_range=feature_range, features=features, prefix=prefix).items() if value is not traitlets.Undefined}

    super(vaex.ml.transformations.MinMaxScaler, self).__init__(**given_kwargs)

MinMaxScaler.__init__ = __init__
MinMaxScaler.__signature__ = __init__
del __init__
    


from ..vaex.ml.transformations import *
import traitlets

def max_abs_scaler(self, features, prefix='absmax_scaled_'):
    obj = vaex.ml.transformations.MaxAbsScaler([{'name': 'features', 'help': 'List of features to transform.'}, {'name': 'prefix', 'help': 'Prefix for the names of the transformed features.'}])
    obj.fit(self)
    return obj

import vaex.dataset
vaex.dataset.Dataset.ml._add(max_abs_scaler=max_abs_scaler)

def __init__(self, features=traitlets.Undefined, prefix='absmax_scaled_'):
    """
    

:param features: List of features to transform.
:param prefix: Prefix for the names of the transformed features.
    """
    given_kwargs = {key:value for key, value in dict(features=features, prefix=prefix).items() if value is not traitlets.Undefined}

    super(vaex.ml.transformations.MaxAbsScaler, self).__init__(**given_kwargs)

MaxAbsScaler.__init__ = __init__
MaxAbsScaler.__signature__ = __init__
del __init__
    


from ..vaex.ml.transformations import *
import traitlets

def robust_scaler(self, features, percentile_range, prefix='robust_scaled_', with_centering=True, with_scaling=True):
    obj = vaex.ml.transformations.RobustScaler([{'name': 'features', 'help': 'List of features to transform.'}, {'name': 'percentile_range', 'help': 'The percentile range to which to scale each feature to.'}, {'name': 'prefix', 'help': 'Prefix for the names of the transformed features.'}, {'name': 'with_centering', 'help': 'If True, remove the median.'}, {'name': 'with_scaling', 'help': 'If True, scale each feature between the specified percentile range.'}])
    obj.fit(self)
    return obj

import vaex.dataset
vaex.dataset.Dataset.ml._add(robust_scaler=robust_scaler)

def __init__(self, features=traitlets.Undefined, percentile_range=traitlets.Undefined, prefix='robust_scaled_', with_centering=True, with_scaling=True):
    """
    

:param features: List of features to transform.
:param percentile_range: The percentile range to which to scale each feature to.
:param prefix: Prefix for the names of the transformed features.
:param with_centering: If True, remove the median.
:param with_scaling: If True, scale each feature between the specified percentile range.
    """
    given_kwargs = {key:value for key, value in dict(features=features, percentile_range=percentile_range, prefix=prefix, with_centering=with_centering, with_scaling=with_scaling).items() if value is not traitlets.Undefined}

    super(vaex.ml.transformations.RobustScaler, self).__init__(**given_kwargs)

RobustScaler.__init__ = __init__
RobustScaler.__signature__ = __init__
del __init__
    


import warnings

import vaex
import vaex.dataset
from . import datasets
from .ui import Widgetize
from .pipeline import Pipeline

from vaex.utils import InnerNamespace

def pca(self, n_components=2, features=None, progress=False):
    '''Requires vaex.ml: Create :class:`vaex.ml.transformations.PCA` and fit it'''
    features = features or self.get_column_names()
    features = vaex.dataset._ensure_strings_from_expressions(features)
    pca = PCA(n_components=n_components, features=features)
    pca.fit(self, progress=progress)
    return pca


def label_encoder(self, features=None, prefix='label_encoded_'):
    '''Requires vaex.ml: Create :class:`vaex.ml.transformations.LabelEncoder` and fit it.'''
    features = features or self.get_column_names()
    features = vaex.dataset._ensure_strings_from_expressions(features)
    label_encoder = LabelEncoder(features=features, prefix=prefix)
    label_encoder.fit(self)
    return label_encoder


def one_hot_encoder(self, features=None, one=1, zero=0, prefix=''):
    '''
    Requires vaex.ml: Create :class:`vaex.ml.transformations.OneHotEncoder` and fit it.

    :param features: list of features to one-hot encode
    :param one: what value to use instead of "1"
    :param zero: what value to use instead of "0"
    :returns one_hot_encoder: vaex.ml.transformations.OneHotEncoder object
    '''
    if features is None:
        raise ValueError('Please give at least one categorical feature.')
    features = vaex.dataset._ensure_strings_from_expressions(features)
    one_hot_encoder = OneHotEncoder(features=features, one=one, zero=zero, prefix=prefix)
    one_hot_encoder.fit(self)
    return one_hot_encoder


def standard_scaler(self, features=None, with_mean=True, with_std=True):
    '''Requires vaex.ml: Create :class:`vaex.ml.transformations.StandardScaler` and fit it'''
    features = features or self.get_column_names()
    features = vaex.dataset._ensure_strings_from_expressions(features)
    standard_scaler = StandardScaler(features=features, with_mean=with_mean, with_std=with_std)
    standard_scaler.fit(self)
    return standard_scaler


def minmax_scaler(self, features=None, feature_range=[0, 1]):
    '''Requires vaex.ml: Create :class:`vaex.ml.transformations.MinMaxScaler` and fit it'''
    features = features or self.get_column_names()
    features = vaex.dataset._ensure_strings_from_expressions(features)
    minmax_scaler = MinMaxScaler(features=features, feature_range=feature_range)
    minmax_scaler.fit(self)
    return minmax_scaler


def xgboost_model(self, label, num_round, features=None, copy=False, param={},
                  prediction_name='xgboost_prediction'):
    '''Requires vaex.ml: create a XGBoost model and train/fit it.

    :param label: label to train/fit on
    :param num_round: number of rounds
    :param features: list of features to train on
    :param bool copy: Copy data or use the modified xgboost library for efficient transfer
    :return vaex.ml.xgboost.XGBModel: fitted XGBoost model
    '''
    from .xgboost import XGBModel
    dataset = self
    features = features or self.get_column_names()
    features = vaex.dataset._ensure_strings_from_expressions(features)
    xg = XGBModel(prediction_name=prediction_name,
                  num_round=num_round,
                  features=features,
                  param=param)
    xg.fit(dataset, label, copy=copy)
    return xg


def lightgbm_model(self, label, num_round, features=None, copy=False, param={},
                   classifier=False, prediction_name='lightgbm_prediction'):
    '''Requires vaex.ml: create a lightgbm model and train/fit it.

    :param label: label to train/fit on
    :param num_round: number of rounds
    :param features: list of features to train on
    :param bool copy: Copy data or use the modified xgboost library for efficient transfer
    :param bool classifier: If true, return a the classifier (will use argmax on the probabilities)
    :return vaex.ml.lightgbm.LightGBMModel or LightGBMClassifier: fitted LightGBM model
    '''
    from .lightgbm import LightGBMModel, LightGBMClassifier
    dataset = self
    features = features or self.get_column_names()
    features = vaex.dataset._ensure_strings_from_expressions(features)
    cls = LightGBMClassifier if classifier else LightGBMModel
    b = cls(prediction_name=prediction_name,
            num_round=num_round,
            features=features,
            param=param)
    b.fit(dataset, label, copy=copy)
    return b

def pygbm_model(self, label, max_iter, features=None, param={}, classifier=False, prediction_name='pygbm_prediction', **kwargs):
    '''Requires vaex.ml: create a pygbm model and train/fit it.

    :param label: label to train/fit on
    :param max_iter: max number of iterations/trees
    :param features: list of features to train on
    :param bool classifier: If true, return a the classifier (will use argmax on the probabilities)
    :return vaex.ml.pygbm.PyGBMModel or vaex.ml.pygbm.PyGBMClassifier: fitted PyGBM model
    '''
    from .incubator.pygbm import PyGBMModel, PyGBMClassifier
    dataset = self
    features = features or self.get_column_names()
    features = vaex.dataset._ensure_strings_from_expressions(features)
    cls = PyGBMClassifier if classifier else PyGBMModel
    b = cls(prediction_name=prediction_name, max_iter=max_iter, features=features, param=param, **kwargs)
    b.fit(dataset, label)
    return b


def state_transfer(self):
    from .transformations import StateTransfer
    state = self.state_get()
    state.pop('active_range')  # we are not interested in this..
    return StateTransfer(state=state)


def train_test_split(self, test_size=0.2, strings=True, virtual=True, verbose=True):
    """Will split the dataset in train and test part, assuming it is shuffled.
    """
    if verbose:
        warnings.warn('make sure the dataset is shuffled')
    initial = None
    try:
        assert self.filtered is False, 'filtered dataset not supported yet'
        # full_length = len(self)
        self = self.trim()
        initial = self.get_active_range()
        self.set_active_fraction(test_size)
        test = self.trim()
        __, end = self.get_active_range()
        self.set_active_range(end, self.length_original())
        train = self.trim()
    finally:
        if initial is not None:
            self.set_active_range(*initial)
    return train, test


def add_namespace():
    vaex.dataset.Dataset.ml = InnerNamespace({}, vaex.dataset.Dataset, prefix='ml_')
    # try:
    #     from . import generated
    # except ImportError:
    #     print("importing generated code failed")
    vaex.dataset.Dataset.ml._add(train_test_split=train_test_split)

    def to_xgboost_dmatrix(self, label, features=None, selection=None, blocksize=1000 * 1000):
        """

        label: ndarray containing the labels
        """
        from . import xgboost
        return xgboost.VaexDMatrix(self, label, features=features, selection=selection, blocksize=blocksize)

    vaex.dataset.Dataset.ml._add(to_xgboost_dmatrix=to_xgboost_dmatrix,
                                 xgboost_model=xgboost_model,
                                 lightgbm_model=lightgbm_model,
                                 pygbm_model=pygbm_model,
                                 state_transfer=state_transfer,
                                 one_hot_encoder=one_hot_encoder,
                                 label_encoder=label_encoder,
                                 pca=pca,
                                 standard_scaler=standard_scaler,
                                 minmax_scaler=minmax_scaler)
    del to_xgboost_dmatrix
add_namespace()
    # named_objects.update({ep.name: ep.load()})
from .transformations import PCA
from .transformations import StandardScaler, MinMaxScaler, MaxAbsScaler, RobustScaler
from .transformations import LabelEncoder, OneHotEncoder



import os
import re
import sys
import traitlets
from . import generate
from collections import defaultdict

# registry = {}
registry = []



def register(cls):
    registry.append(cls)
    # registry[(cls.__module__, cls.__name__)] = cls
    return cls

def camel_to_underscore(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()

def generate_vaex_ml():
    from jinja2 import Template
    import vaex.ml.transformations
    import vaex.ml.cluster
    import vaex.ml.lightgbm
    import vaex.ml.xgboost
    import vaex.ml.linear_model
    try:
        import vaex.ml.incubator.annoy
    except:
        pass
    try:
        import vaex.ml.incubator.pygbm
    except:
        pass

    docstring_args_template = Template("""
{% for arg in args %}
:param {{ arg.name }}: {{ arg.help }}{% endfor %}
""")

    template_method = Template("""

from ..{{ module }} import *
import traitlets

def {{ method_name }}(self, {{ signature }}):
    obj = {{ module }}.{{ class_name }}({{ args }})
    obj.fit(self{{ extra_fit }})
    return obj

import vaex.dataset
vaex.dataset.Dataset.ml._add({{ method_name }}={{ method_name }})

def __init__(self, {{ full_signature }}):
    \"\"\"
    {{ docstring_args }}
    \"\"\"
    given_kwargs = {key:value for key, value in dict({{ full_args }}).items() if value is not traitlets.Undefined}

    super({{ module }}.{{ class_name }}, self).__init__(**given_kwargs)

{{ class_name }}.__init__ = __init__
{{ class_name }}.__signature__ = __init__
del __init__
    """)
    filename = os.path.join(os.path.dirname(__file__), "generated.py")
    snippets = defaultdict(list) # module name -> list of snippets
    # with open(filename, "w") as f:
    for cls in generate.registry:
        traits = {key: value for key, value in cls.class_traits().items()
                    if 'output' not in value.metadata}
        traits_nodefault = {key: value for key, value in traits.items()
                            if value.default_value is traitlets.Undefined}
        import IPython
        IPython
        traits_default = {key: value for key, value in traits.items()
                            if key not in traits_nodefault}
        signature_list = ["{name}".format(name=name, value=value.default_value)
                            for name, value in traits_nodefault.items()]
        signature_list.extend(["{name}={value!r}".format(name=name, value=value.default_value)
                                for name, value in traits_default.items()])
        signature = ", ".join(signature_list)
        args = ", ".join(["{name}={name}".format(name=name, value=value)
                            for name, value in traits.items()])

        signature_list = ["{name}={value!r}".format(name=name, value=value.default_value)
                            for name, value in traits.items()]
        full_signature = ", ".join(signature_list)
        full_args = args

        method_name = camel_to_underscore(cls.__name__)
        module = cls.__module__
        class_name = cls.__name__

        doctraits = {name: trait for name, trait in traits.items() if 'help' in trait.metadata}
        args = [{'name': name, 'help': trait.metadata['help']} for name, trait in doctraits.items()]

        kwargs = dict(locals())
        docstring_args = docstring_args_template.render(**kwargs)

        kwargs = dict(locals())
        code = template_method.render(**kwargs)
        snippets[cls.__module__].append(code)
    for module_name, snippets in snippets.items():
        path = sys.modules[module_name].__file__
        basepath, dot, ext = path.rpartition('.')
        dirname, basename = os.path.dirname(basepath), os.path.basename(basepath)
        gen_path = os.path.join(dirname, 'autogen', basename + '.py')
        print(gen_path)
        with open(gen_path, "w") as f:
            f.write('\n'.join(snippets))


def main(argv=sys.argv):
    generate_vaex_ml()


if __name__ == '__main__':
    main()





from ..vaex.ml.lightgbm import *
import traitlets

def light_gbm_model(self, features, param, num_round=0, prediction_name='lightgbm_prediction'):
    obj = vaex.ml.lightgbm.LightGBMModel([{'name': 'features', 'help': 'List of features to use when fitting the LightGBMModel.'}, {'name': 'num_round', 'help': 'Number of boosting iterations.'}, {'name': 'param', 'help': 'parameters to be passed on the to the LightGBM model.'}, {'name': 'prediction_name', 'help': 'The name of the virtual column housing the predictions.'}])
    obj.fit(self)
    return obj

import vaex.dataset
vaex.dataset.Dataset.ml._add(light_gbm_model=light_gbm_model)

def __init__(self, features=traitlets.Undefined, num_round=0, param=traitlets.Undefined, prediction_name='lightgbm_prediction'):
    """
    

:param features: List of features to use when fitting the LightGBMModel.
:param num_round: Number of boosting iterations.
:param param: parameters to be passed on the to the LightGBM model.
:param prediction_name: The name of the virtual column housing the predictions.
    """
    given_kwargs = {key:value for key, value in dict(features=features, num_round=num_round, param=param, prediction_name=prediction_name).items() if value is not traitlets.Undefined}

    super(vaex.ml.lightgbm.LightGBMModel, self).__init__(**given_kwargs)

LightGBMModel.__init__ = __init__
LightGBMModel.__signature__ = __init__
del __init__
    


from ..vaex.ml.lightgbm import *
import traitlets

def light_gbm_classifier(self, features, param, num_round=0, prediction_name='lightgbm_prediction'):
    obj = vaex.ml.lightgbm.LightGBMClassifier([{'name': 'features', 'help': 'List of features to use when fitting the LightGBMModel.'}, {'name': 'num_round', 'help': 'Number of boosting iterations.'}, {'name': 'param', 'help': 'parameters to be passed on the to the LightGBM model.'}, {'name': 'prediction_name', 'help': 'The name of the virtual column housing the predictions.'}])
    obj.fit(self)
    return obj

import vaex.dataset
vaex.dataset.Dataset.ml._add(light_gbm_classifier=light_gbm_classifier)

def __init__(self, features=traitlets.Undefined, num_round=0, param=traitlets.Undefined, prediction_name='lightgbm_prediction'):
    """
    

:param features: List of features to use when fitting the LightGBMModel.
:param num_round: Number of boosting iterations.
:param param: parameters to be passed on the to the LightGBM model.
:param prediction_name: The name of the virtual column housing the predictions.
    """
    given_kwargs = {key:value for key, value in dict(features=features, num_round=num_round, param=param, prediction_name=prediction_name).items() if value is not traitlets.Undefined}

    super(vaex.ml.lightgbm.LightGBMClassifier, self).__init__(**given_kwargs)

LightGBMClassifier.__init__ = __init__
LightGBMClassifier.__signature__ = __init__
del __init__
    




from ..vaex.ml.xgboost import *
import traitlets

def xgb_model(self, features, param, num_round=0, prediction_name='xgboost_prediction'):
    obj = vaex.ml.xgboost.XGBModel([])
    obj.fit(self)
    return obj

import vaex.dataset
vaex.dataset.Dataset.ml._add(xgb_model=xgb_model)

def __init__(self, features=traitlets.Undefined, num_round=0, param=traitlets.Undefined, prediction_name='xgboost_prediction'):
    """
    

    """
    given_kwargs = {key:value for key, value in dict(features=features, num_round=num_round, param=param, prediction_name=prediction_name).items() if value is not traitlets.Undefined}

    super(vaex.ml.xgboost.XGBModel, self).__init__(**given_kwargs)

XGBModel.__init__ = __init__
XGBModel.__signature__ = __init__
del __init__
    


import os
import ctypes
import math

import vaex
import lightgbm
import numpy as np
import tempfile
import base64
import vaex.serialize
from . import state
import traitlets

from . import generate

#libpath = os.path.join(os.path.dirname(__file__), "vaex_ml_lib.cpython-35m-x86_64-linux-gnu.so")
#print(libpath)
#lib = ctypes.cdll.LoadLibrary(libpath)
lib = lightgbm.basic._LIB

class EXPRESSION_RESULT(ctypes.Structure):
     _fields_ = [
         ("length", ctypes.c_longlong),
         ("data", ctypes.POINTER(ctypes.c_float))
     ]

EVALFUNC = ctypes.CFUNCTYPE(ctypes.c_int, (ctypes.c_char_p), ctypes.c_longlong, ctypes.c_longlong, ctypes.POINTER(EXPRESSION_RESULT))

@vaex.serialize.register
@generate.register
class LightGBMModel(state.HasState):
    '''The LightGBM algorithm.

    This class provides an interface to the LightGBM aloritham, with some optimizations
    for better memory efficiency when training large datasets. The algorithm itself is
    not modified at all.

    LightGBM is a fast gradient boosting algorithm based on decision trees and is
    mainly used for classification, regression and ranking tasks. It is under the
    umbrella of the Distributed Machine Learning Toolkit (DMTK) project of Microsoft.
    For more information, please visit https://github.com/Microsoft/LightGBM/.

    Example:
    --------

    import vaex.ml.
    >>> import vaex.ml.lightgbm
    >>> df = vaex.ml.datasets.load_iris()
    >>> features = ['sepal_width', 'petal_length', 'sepal_length', 'petal_width']
    >>> df_train, df_test = vaex.ml.train_test_split(df)
    >>> params = {
        'boosting': 'gbdt',
        'max_depth': 5,
        'learning_rate': 0.1,
        'application': 'multiclass',
        'num_class': 3,
        'subsample': 0.80,
        'colsample_bytree': 0.80}
    >>> booster = vaex.ml.lightgbm.LightGBMModel(features=features, num_rounds=100, param=params)
    >>> booster.fit(df_train, 'class_')
    >>> df_train = booster.transform(df_train)
    >>> df_test = booster.transform(df_test)
    '''

    features = traitlets.List(traitlets.Unicode(), help='List of features to use when fitting the LightGBMModel.')
    num_round = traitlets.CInt(help='Number of boosting iterations.')
    param = traitlets.Dict(help='parameters to be passed on the to the LightGBM model.')
    prediction_name = traitlets.Unicode(default_value='lightgbm_prediction', help='The name of the virtual column housing the predictions.')


    def __call__(self, *args):
        data2d = np.vstack([arg.astype(np.float64) for arg in args]).T.copy()
        #dmatrix = lightgbm.Dataset(data2d)
        return self.bst.predict(data2d)

    def transform(self, dataset):
        '''
        Transform the dataset such that it contains the predictions of the LightGBMModel
        in a form of a virtual columns.

        Parameters:
        -----------
        :param dataset: A vaex dataset.

        Returns:
        --------
        :return copy: A shallow copy of the dataset that includes the LightGBMModel predictions as virtual columns.
        '''

        copy = dataset.copy()
        lazy_function = copy.add_function('lightgbm_prediction_function', self)
        expression = lazy_function(*self.features)
        copy.add_virtual_column(self.prediction_name, expression, unique=False)
        return copy

    def fit(self, dataset, label, copy=False):
        '''
        Fit the LightGBMModel to the dataset.

        Parameters:
        -----------
        :param dataset: A vaex dataset.
        :param label: The name of the column containing the target variable.
        :param copy: bool, if True, make an in memory copy of the data before passing it to the LightGBMModel.

        Returns:
        --------
        self
        '''
        if copy:
            data = np.vstack([dataset.evaluate(k).astype(np.float64) for k in self.features]).T
            label_data = dataset.evaluate(label)
            dtrain = lightgbm.Dataset(data, label_data)
        else:
            dtrain = VaexDataset(dataset, label, features=self.features)
        param = dict(self.param)
        self.bst = lightgbm.train(param, dtrain, self.num_round)

    def predict(self, dataset, copy=False):
        '''Get an in-memory numpy array with the predictions of the LightGBMModel on
        a vaex dataset

        Parameters:
        -----------
        :param dataset: A vaex dataset.
        :param copy: bool, if True, make an in memory copy of the data before passing it to the LightGBMModel.

        Returns:
        --------
        A in-memory numpy array containing the LightGBMModel predictions.
        '''

        # TODO: we want to go multithreaded/parallel/chunks
        data = np.vstack([dataset.evaluate(k) for k in self.features]).T
        # if copy:
        #   data = np.vstack([dataset.evaluate(k) for k in self.features]).T
        #   dmatrix = lightgbm.Dataset(data)
        # else:
        #   dmatrix = VaexDMatrix(dataset, features=self.features)
        return self.bst.predict(data)

    def state_get(self):
        filename = tempfile.mktemp()
        self.bst.save_model(filename)
        with open(filename, 'rb') as f:
            data = f.read()
        return dict(tree_state=base64.encodebytes(data).decode('ascii'),
            substate=super(LightGBMModel, self).state_get())

    def state_set(self, state):
        super(LightGBMModel, self).state_set(state['substate'])
        data = base64.decodebytes(state['tree_state'].encode('ascii'))
        filename = tempfile.mktemp()
        with open(filename, 'wb') as f:
            f.write(data)
        self.bst = lightgbm.Booster(model_file=filename)

@vaex.serialize.register
@generate.register
class LightGBMClassifier(LightGBMModel):
    def __call__(self, *args):
        return np.argmax(super(LightGBMClassifier, self).__call__(*args), axis=1)
    def predict(self, dataset, copy=False):
        return np.argmax(super(LightGBMClassifier, self).predict(dataset, copy=copy), axis=1)


class VaexDataset(lightgbm.Dataset):
    def __init__(self, ds, label=None, features=None, blocksize=100*1000, sample_count=10*100, params={}):
        super(VaexDataset, self).__init__(None)
        self.ds = ds
        self.features = features or self.ds.get_column_names(virtual=True)
        assert len(set(self.features)) == len(self.features), "using duplicate features"
        self.blocksize = blocksize

        self.data_references = {}

        self.c_features = (ctypes.c_char_p * len(self.features))()
        self.c_features[:] = [k.encode() for k in self.features]

        self.c_evaluate = EVALFUNC(self.evaluate)

        self.handle = ctypes.c_void_p()

        row_count = len(self.ds)
        ncol = len(self.features)
        num_sample_row = min(sample_count, row_count)

        data_list = [self.ds.evaluate(k, i1=0, i2=num_sample_row).astype(np.float64) for k in self.features]
        data_pointers = [k.ctypes.data_as(ctypes.POINTER(ctypes.c_double)) for k in data_list]
        sample_data = (ctypes.POINTER(ctypes.c_double) * ncol)(*data_pointers)

        indices = np.arange(num_sample_row, dtype=np.int32)
        indices_pointers = [indices.ctypes.data_as(ctypes.POINTER(ctypes.c_int)) for k in range(ncol)]
        sample_indices = (ctypes.POINTER(ctypes.c_int) * ncol)(*indices_pointers)
        num_per_col = (ctypes.c_int * ncol)(*((num_sample_row,) * ncol))
        parameters = ctypes.c_char_p(lightgbm.basic.param_dict_to_str(params).encode())

        lightgbm.basic._safe_call(lib.LGBM_DatasetCreateFromSampledColumn(sample_data,
                                                sample_indices,
                                                ctypes.c_uint(ncol),
                                                num_per_col,
                                                ctypes.c_uint(num_sample_row),
                                                ctypes.c_uint(row_count),
                                                parameters,
                                                ctypes.byref(self.handle)))
        #self._push_rows()
        #print(">>>", self.num_data())
        blocks = int(math.ceil(row_count / blocksize))
        #print(blocks)
        dtype = np.float64
        for i in range(blocks):
            i1 = i * blocksize
            i2 = min(row_count, (i+1) * blocksize)
            data = np.array([ds.evaluate(k, i1=i1, i2=i2).astype(dtype) for k in features]).T.copy()
            #print(data.shape)
            ctypemap = {np.float64: ctypes.c_double, np.float32: ctypes.c_float}
            capi_typemap = {np.float64: lightgbm.basic.C_API_DTYPE_FLOAT64, np.float32:lightgbm.basic.C_API_DTYPE_FLOAT32}
            lightgbm.basic._safe_call(lib.LGBM_DatasetPushRows(self.handle,
                                     data.ctypes.data_as(ctypes.POINTER(ctypemap[dtype])),
                                     ctypes.c_uint(capi_typemap[dtype]),
                                     ctypes.c_uint32(i2-i1),
                                     ctypes.c_uint32(ncol),
                                     ctypes.c_uint32(i1),
                                     ))

        #return_code = lib.create_dmatrix(self.c_features, len(self.features), rows, batch_size, self.c_evaluate, self.blocksize, ctypes.byref(self.handle))
        #print(return_code)
        ##lib.test_dmatrix(self.handle, 0)
        if label is not None:
            self.label_data = self.ds.evaluate(label)
            #$self.set_label_npy2d(label)
            self.set_label(self.label_data)


    def set_label_npy2d(self, label):
        """Set label of dmatrix

        Parameters
        ----------
        label: array like
            The label information to be set into DMatrix
            from numpy 2D array
        """
        self.set_float_info_npy2d('label', label)

    def set_float_info_npy2d(self, field, data):
        """Set float type property into the DMatrix
           for numpy 2d array input

        Parameters
        ----------
        field: str
            The field name of the information

        data: numpy array
            The array of data to be set
        """
        data = np.array(data, copy=False, dtype=np.float32)
        c_data = data.ctypes.data_as(ctypes.POINTER(ctypes.c_float))
        lightgbm.core._check_call(lightgbm.core._LIB.XGDMatrixSetFloatInfo(self.handle,
                                               lightgbm.core.c_str(field),
                                               c_data,
                                               ctypes.c_uint64(len(data))))

    def evaluate(self, expression, i1, i2, result):
        try:
            expression = expression.decode()
            #print(expression, i1, i2)
            #print(result[0])
            data = self.ds.evaluate(expression, i1, i2) # TODO, support selections
            # keep a reference, since otherwise the memory is freed
            # and we overwrite the last reference, so that will be freed
            self.data_references[expression] = data = data.astype(np.float32) # TODO: think about masking, nan?
            #print(data, type(data), data.dtype)
            ptr = data.ctypes.data_as(ctypes.POINTER(ctypes.c_float))
            #print(ptr, ptr)
            result[0].length = len(data)
            result[0].data = ptr
            #print(result.data, result.length)
            #print(result)
            return 0 #EXPRESSION_RESULT(len(data), ptr)
        except (Exception, e):
            print('error', e)
try:
    from .autogen import lightgbm
    del lightgbm
except ImportError:
    pass

if __name__ == "__main__":
    ds = vaex.ml.iris()
    features = ['sepal_length', 'sepal_width', 'petal_length', 'petal_width']
    #lds = VaexDataset(ds, 'class_', features=features)
    lds = lightgbm.Dataset(np.array(ds[features]), ds.data.class_)
    param = {'num_leaves':31, 'num_trees':100, 'objective':'softmax', 'num_class': 3}
    num_round=30
    bst = lightgbm.train(param, lds, num_round)
    print(bst.predict(np.array(ds[features])))

